open Ptree
open Pitptree
open Parsing_helper

(* Module Set and Map for string *)

module CompString =
  struct
    type t = string
    let compare  = Pervasives.compare
  end

module StringMap = Map.Make(CompString)

(* Special names *)

let stamp_str = "st"
let nat_str = "i"
let priv_channel_str = "d"
let query_var_str = "x"
let query_name_str = "n"

(* Options *)
let ignore_type = ref false
let replace_types_by_bitstring = ref false
let allowDiffPatterns = ref false
let reconstructTrace = ref true


(***********************************
          Initialisation
************************************)

(* Initialise environment *)

let any_type_id = { id = "_any_type"; counter = 0 }
let bitstring_id = { id = "bitstring"; counter = 0 }
let channel_id = { id = "channel"; counter = 0 }
let bool_id = { id = "bool"; counter = 0 }
let nat_id = { id = "nat"; counter = 0 }
let stamp_id = { id = "stamp"; counter = 0 }

let true_id = { id = "true"; counter = 0 }
let false_id = { id = "false"; counter = 0 }

let error_v = {id_v = {id="error"; counter=0}; typ_v=bitstring_id; status_v=SCNone; options_v=[]; side_v = NoRelevant}

(* Operators *)

let and_id = { id = "&&"; counter = 0 }
let or_id = { id = "||"; counter = 0 }
let eq_id = { id = "="; counter = 0 }
let neq_id = { id = "<>"; counter = 0 }
let inf_id = { id = "<"; counter = 0 }
let infeq_id = { id = "<="; counter = 0 }
let not_id = { id = "not"; counter = 0 }
let choice_id = { id = "choice"; counter = 0 }

let and_symb = { id_f = and_id; arg_ty = [bool_id; bool_id]; return_ty = bool_id; data = false; priv = false; type_conv = false; pure_priv = false; funcat = Logic }
let or_symb = { id_f = or_id; arg_ty = [bool_id; bool_id]; return_ty = bool_id; data = false; priv = false; type_conv = false; pure_priv = false; funcat = Logic}
let eq_symb = { id_f = eq_id; arg_ty = [any_type_id; any_type_id]; return_ty = bool_id; data = false; priv = false; type_conv = false; pure_priv = false; funcat = Logic }
let neq_symb = { id_f = neq_id; arg_ty = [any_type_id; any_type_id]; return_ty = bool_id; data = false; priv = false; type_conv = false; pure_priv = false; funcat = Logic }
let inf_symb = { id_f = inf_id; arg_ty = [nat_id; nat_id]; return_ty = bool_id; data = false; priv = false; type_conv = false; pure_priv = false; funcat = Logic }
let infeq_symb = { id_f = infeq_id; arg_ty = [nat_id; nat_id]; return_ty = bool_id; data = false; priv = false; type_conv = false; pure_priv = false; funcat = Logic }
let not_symb = { id_f = not_id; arg_ty = [bool_id]; return_ty = bool_id; data = false; priv = false; type_conv = false; pure_priv = false; funcat = Logic }
let true_symb = { id_f = true_id; arg_ty = []; return_ty = bool_id; data = false; priv = false; type_conv = false; pure_priv = false; funcat = Logic }
let false_symb = { id_f = false_id; arg_ty = []; return_ty = bool_id; data = false; priv = false; type_conv = false; pure_priv = false; funcat = Logic }
let choice_symb = { id_f = choice_id; arg_ty = [bitstring_id; bitstring_id]; return_ty = bitstring_id; data = false; priv = false; type_conv = false; pure_priv = false; funcat = Choice }

(* constants *)
let badL = { id_n = { id = "badL"; counter = 0 }; typ_n = bitstring_id; options_n = []; ext_n = dummy_ext; status_n = SCNone }
let badR = { id_n = { id = "badR"; counter = 0 }; typ_n = bitstring_id; options_n = []; ext_n = dummy_ext; status_n = SCNone }
let c_error = { id_n = { id = "c_error"; counter = 0 }; typ_n = channel_id; options_n = []; ext_n = dummy_ext; status_n = SCPublic }

let initial_environment =
  let env1 = StringMap.add "bitstring" (EType bitstring_id) StringMap.empty in
  let env2 = StringMap.add "channel" (EType channel_id) env1 in
  let env3 = StringMap.add "bool" (EType bool_id) env2 in
  let env4 = StringMap.add "nat" (EType nat_id) env3 in
  let env5 = StringMap.add "true" (EFun true_symb) env4 in
  let env6 = StringMap.add "false" (EFun false_symb) env5 in
  let env7 = StringMap.add "&&" (EFun and_symb) env6 in
  let env8 = StringMap.add "||" (EFun or_symb) env7 in
  let env9 = StringMap.add "=" (EFun eq_symb) env8 in
  let env10 = StringMap.add "<>" (EFun neq_symb) env9 in
  let env11 = StringMap.add "<" (EFun inf_symb) env10 in
  let env12 = StringMap.add "<=" (EFun infeq_symb) env11 in
  let env13 = StringMap.add "not" (EFun not_symb) env12 in
  let env14 = StringMap.add "choice" (EFun choice_symb) env13 in
  env14

(* Freshness *)

let fresh_id_from_ident, fresh_id_from_string =
  let acc = ref 1 in
  let fresh_ident (str,_) =
    let r = { id = str; counter = !acc } in
    incr acc;
    r
  in
  let fresh_string str =
    let r = { id = str; counter = !acc } in
    incr acc;
    r
  in
  fresh_ident, fresh_string

let fresh_var_nat side () =
  let n = fresh_id_from_string nat_str in
  { id_v = n; typ_v = nat_id; status_v = SCNone; options_v = []; side_v = side }

let fresh_var_stamp side () =
  let n = fresh_id_from_string stamp_str in
  { id_v = n; typ_v = stamp_id; status_v = SCNone; options_v = []; side_v = side }

let fresh_name_stamp () =
  let n = fresh_id_from_string stamp_str in
  { id_n = n; typ_n = stamp_id; options_n = []; ext_n = dummy_ext; status_n = SCNone }

let fresh_name_channel () =
  let n = fresh_id_from_string priv_channel_str in
  { id_n = n; typ_n = channel_id; options_n = []; ext_n = dummy_ext; status_n = SCNone }

let fresh_var typ side =
  if typ = nat_id
  then fresh_var_nat side ()
  else if typ = stamp_id
  then fresh_var_stamp side ()
  else if typ = channel_id
  then
    let n = fresh_id_from_string priv_channel_str in
    { id_v = n; typ_v = typ; status_v = SCNone; options_v = []; side_v = side }
  else
    let n = fresh_id_from_string query_var_str in
    { id_v = n; typ_v = typ; status_v = SCNone; options_v = []; side_v = side }

let fresh_name typ =
  if typ = channel_id
  then fresh_name_channel ()
  else if typ = stamp_id
  then fresh_name_stamp ()
  else if typ = nat_id
  then
    failwith "[transform.ml >> fresh_name] Fresh names cannot be of type Nat"
  else
    let n = fresh_id_from_string query_name_str in
    { id_n = n; typ_n = typ; options_n = []; ext_n = dummy_ext; status_n = SCNone }


(* Retreive from environement *)
let get_variable_from_id env (str,ext) =
  try
    match StringMap.find str env with
      | EVar var -> var
      | _ -> input_error (Printf.sprintf "The identifier %s is expected to be a variable." str) ext
  with
  | Not_found -> input_error (Printf.sprintf "The identifier %s is not defined." str) ext

let get_type_from_id env (str,ext) =
  try
    match StringMap.find str env with
      | EType id -> id
      | _ -> input_error (Printf.sprintf "The identifier %s is expected to be a type." str) ext
  with
  | Not_found -> input_error (Printf.sprintf "The identifier %s is not defined." str) ext

let get_process_from_id env (str,ext) =
  try
    match StringMap.find str env with
      | EProcess (id,ty_l,gen_proc) -> (id,ty_l,gen_proc)
      | _ -> input_error (Printf.sprintf "The identifier %s is expected to be a process." str) ext
  with
  | Not_found -> input_error (Printf.sprintf "The identifier %s is not defined." str) ext

let get_function_from_id env (str,ext) =
  try
    match StringMap.find str env with
      | EFun f -> f
      | _ -> input_error (Printf.sprintf "The identifier %s is expected to be a function." str) ext
  with
  | Not_found -> input_error (Printf.sprintf "The identifier %s is not defined." str) ext

let get_event_from_id env (str,ext) =
  try
    match StringMap.find str env with
      | EEvent ev -> ev
      | _ -> input_error (Printf.sprintf "The identifier %s is expected to be an event." str) ext
  with
  | Not_found -> input_error (Printf.sprintf "The identifier %s is not defined." str) ext

let get_table_from_id env (str,ext) =
  try
    match StringMap.find str env with
      | ETable tbl -> tbl
      | _ -> input_error (Printf.sprintf "The identifier %s is expected to be a table." str) ext
  with
  | Not_found -> input_error (Printf.sprintf "The identifier %s is not defined." str) ext

(* Errors *)

let display_list f_elt sep = function
  | [] -> ""
  | [t] -> f_elt t
  | t::q -> List.fold_left (fun acc t' -> Printf.sprintf "%s%s%s" acc sep (f_elt t')) (f_elt t) q

(* Type comparaison *)

let is_equal_type typ1 typ2 =
  (typ1 == any_type_id) ||
  (typ2 == any_type_id) ||
  (typ1 == nat_id && typ2 == nat_id) ||
  (not !ignore_type && typ1 == typ2) ||
  (!ignore_type && typ1 != nat_id && typ2 != nat_id)

let is_equal_type_list typl1 typl2 = List.for_all2 is_equal_type typl1 typl2

let rec get_type = function
  | Var v -> v.typ_v
  | DumVar v -> v.typ_v
  | Name n -> n.typ_n
  | Sum _ -> nat_id
  | Nat _ -> nat_id
  | Fun(f,_) -> f.return_ty
  | Tuple _ -> bitstring_id
  | Restr(_,_,t)
  | Test(_,t,_)
  | Let(_,_,t,_)
  | LetFilter(_,_,t,_) -> get_type t

(****************************************
***          Expand protocol          ***
*****************************************)

let restricted_name = ref StringMap.empty

let rec expand_term_e env (t,ext) = match t with
  | PIdent (str,ext) ->
      begin
        try
          match StringMap.find str env with
            | EFun f ->
                if f.arg_ty <> []
                then input_error (Printf.sprintf "The identifier %s is a function that expect %d arguments." str (List.length f.arg_ty)) ext;

                (Fun(f,[]),f.return_ty)
            | EVar v -> (Var v, v.typ_v)
            | EName n -> (Name n, n.typ_n)
            | EArg (t,ty) -> (t,ty)
            | EProcess _ -> input_error (Printf.sprintf "The process %s cannot be called in a term." str) ext
            | ETable _ -> input_error (Printf.sprintf "The table %s cannot appear in a term." str) ext
            | EEvent _ -> input_error (Printf.sprintf "The event %s cannot appear in a term." str) ext
            | _ -> input_error (Printf.sprintf "The type %s cannot appear in a term." str) ext
        with Not_found -> input_error (Printf.sprintf "Identifier %s not defined." str) ext
      end
  | PNat n -> (Nat n, nat_id)
  | PSum ((str,ext),n) ->
      let v = get_variable_from_id env (str,ext) in

      if not (is_equal_type v.typ_v nat_id)
      then input_error (Printf.sprintf "The function + expects arguments of type nat but received an argument of type %s." v.typ_v.id) ext;

      (Sum(v,n), nat_id)
  | PFail -> input_error "The front end does not handle function declaration with the Fail constant." ext
  | PFunApp((str,ext),args) ->
      begin
        try
          match StringMap.find str env with
            | EFun f ->
                let (new_args,new_types) =
                  List.fold_right (fun t (acc_t,acc_ty) ->
                    let (new_t,new_ty) = expand_term_e env t in
                    (new_t::acc_t,new_ty::acc_ty)
                  ) args ([],[])
                in

                if not (is_equal_type_list new_types f.arg_ty)
                then input_error (Printf.sprintf "The function %s is expecting arguments of types %s but received arguments of type %s" f.id_f.id (display_list (fun id -> id.id) ", " f.arg_ty) (display_list (fun id -> id.id) ", " new_types)) ext;

                (Fun (f,new_args),f.return_ty)
            | _ -> input_error (Printf.sprintf "The identifier %s is expected to be a function symbol." str) ext
        with Not_found -> input_error (Printf.sprintf "Identifier %s not defined." str) ext
      end
  | PTuple(args) ->
      let new_args =
        List.fold_right (fun t acc_t ->
          let (new_t,_) = expand_term_e env t in
          new_t::acc_t
        ) args []
      in
      Tuple(new_args),bitstring_id

let rec expand_and_check_pure_priv_term_e env (t,ext) = match t with
  | PIdent (str,ext) ->
      begin
        try
          match StringMap.find str env with
            | EFun f ->
                if f.arg_ty <> []
                then input_error (Printf.sprintf "The identifier %s is a function that expect %d arguments." str (List.length f.arg_ty)) ext;

                (Fun(f,[]),f.return_ty)
            | EVar v -> (Var v, v.typ_v)
            | EName n -> (Name n, n.typ_n)
            | EArg (t,ty) -> (t,ty)
            | EProcess _ -> input_error (Printf.sprintf "The process %s cannot be called in a term." str) ext
            | ETable _ -> input_error (Printf.sprintf "The table %s cannot appear in a term." str) ext
            | EEvent _ -> input_error (Printf.sprintf "The event %s cannot appear in a term." str) ext
            | _ -> input_error (Printf.sprintf "The type %s cannot appear in a term." str) ext
        with Not_found -> input_error (Printf.sprintf "Identifier %s not defined." str) ext
      end
  | PNat n -> (Nat n, nat_id)
  | PSum ((str,ext),n) ->
      let v = get_variable_from_id env (str,ext) in

      if not (is_equal_type v.typ_v nat_id)
      then input_error (Printf.sprintf "The function + expects arguments of type nat but received an argument of type %s." v.typ_v.id) ext;

      (Sum(v,n), nat_id)
  | PFail -> input_error "The front end does not handle function declaration with the Fail constant." ext
  | PFunApp((str,ext),args) ->
      begin
        try
          match StringMap.find str env with
            | EFun f ->
                if f.priv then f.pure_priv <- false;

                let (new_args,new_types) =
                  List.fold_right (fun t (acc_t,acc_ty) ->
                    let (new_t,new_ty) = expand_and_check_pure_priv_term_e env t in
                    (new_t::acc_t,new_ty::acc_ty)
                  ) args ([],[])
                in

                if not (is_equal_type_list new_types f.arg_ty)
                then input_error (Printf.sprintf "The function %s is expenting arguments of types %s but received arguments of type %s" f.id_f.id (display_list (fun id -> id.id) ", " f.arg_ty) (display_list (fun id -> id.id) ", " new_types)) ext;

                (Fun (f,new_args),f.return_ty)
            | _ -> input_error (Printf.sprintf "The identifier %s is expected to be a function symbol." str) ext
        with Not_found -> input_error (Printf.sprintf "Identifier %s not defined." str) ext
      end
  | PTuple(args) ->
      let new_args =
        List.fold_right (fun t acc_t ->
          let (new_t,_) = expand_and_check_pure_priv_term_e env t in
          new_t::acc_t
        ) args []
      in
      Tuple(new_args),bitstring_id

let rec expand_pterm env main_ext = function
  | PPIdent (str,ext) ->
      begin
        try
          match StringMap.find str env with
            | EFun f ->
                if f.arg_ty <> []
                then input_error (Printf.sprintf "The identifier %s is a function that expect %d arguments." str (List.length f.arg_ty)) ext;

                (Fun(f,[]),f.return_ty)
            | EVar v -> (Var v, v.typ_v)
            | EName n -> (Name n, n.typ_n)
            | EArg (t,ty) -> (t,ty)
            | EProcess _ -> input_error (Printf.sprintf "The process %s cannot be called in a term." str) ext
            | ETable _ -> input_error (Printf.sprintf "The table %s cannot appear in a term." str) ext
            | EEvent _ -> input_error (Printf.sprintf "The event %s cannot appear in a term." str) ext
            | _ -> input_error (Printf.sprintf "The type %s cannot appear in a term." str) ext
        with Not_found -> input_error (Printf.sprintf "Identifier %s not defined." str) ext
      end
  | PPNat n -> (Nat n, nat_id)
  | PPSum ((str,ext),n) ->
      let v = get_variable_from_id env (str,ext) in

      if not (is_equal_type v.typ_v nat_id)
      then input_error (Printf.sprintf "The function + expects arguments of type nat but received an argument of type %s." v.typ_v.id) ext;

      (Sum(v,n), nat_id)
  | PPFunApp((str,ext),args) ->
      begin
        try
          match StringMap.find str env with
            | EFun f ->
                let (new_args,new_types) =
                  List.fold_right (fun t (acc_t,acc_ty) ->
                    let (new_t,new_ty) = expand_pterm_e env t in
                    (new_t::acc_t,new_ty::acc_ty)
                  ) args ([],[])
                in

                if not (is_equal_type_list new_types f.arg_ty)
                then input_error (Printf.sprintf "The function %s is expenting arguments of types %s but received arguments of type %s" f.id_f.id (display_list (fun id -> id.id) ", " f.arg_ty) (display_list (fun id -> id.id) ", " new_types)) ext;

                (Fun (f,new_args),f.return_ty)
            | _ -> input_error (Printf.sprintf "The identifier %s is expected to be a function symbol." str) ext
        with Not_found -> input_error (Printf.sprintf "Identifier %s not defined." str) ext
      end
  | PPTuple(args) ->
      let new_args =
        List.fold_right (fun t acc_t ->
          let (new_t,_) = expand_pterm_e env t in
          new_t::acc_t
        ) args []
      in
      Tuple(new_args),bitstring_id
  | PPRestr((str,ext),vars_op,ty, term) ->
      let new_id = fresh_id_from_ident (str,ext) in
      let new_vars_op = match vars_op with
        | None -> None
        | Some vars -> Some (List.map (get_variable_from_id env) vars)
      in
      let new_type = get_type_from_id env ty in
      let new_name = { id_n = new_id; typ_n = new_type; options_n = []; status_n = SCNone; ext_n = ext } in
      let new_env = StringMap.add new_id.id (EName new_name) env in
      let new_term,return_type = expand_pterm_e new_env term in

      Restr(new_name,new_vars_op,new_term), return_type
  | PPTest(cond,t_then,t_op) ->
      let new_cond,cond_type = expand_pterm_e env cond in

      if not (is_equal_type cond_type bool_id)
      then input_error (Printf.sprintf "The condition of the test is expected to be of type bool but is of type %s" cond_type.id) main_ext;

      let new_t_then,t_then_type = expand_pterm_e env t_then in
      let new_t_op = match t_op with
        | None -> None
        | Some t ->
            let new_t, t_type = expand_pterm_e env t in
            if not (is_equal_type t_type t_then_type)
            then input_error (Printf.sprintf "The type of the then branch %s differs from the type of the else branch %s." t_then_type.id t_type.id) main_ext;
            Some new_t
      in
      Test(new_cond,new_t_then,new_t_op),t_then_type
  | PPLet(pat,t,t_then,t_op) ->
      let new_t,_ = expand_pterm_e env t in
      let new_pat, new_env = expand_tpattern env (Some new_t) pat in

      let new_t_then, new_ty_then = expand_pterm_e new_env t_then in
      begin match t_op with
        | None -> Let(new_pat,new_t,new_t_then, None), new_ty_then
        | Some t_else ->
            let new_t_else,new_ty_else = expand_pterm_e env t_else in
            if not (is_equal_type new_ty_then new_ty_else)
            then input_error (Printf.sprintf "The type of the then branch %s differs from the type of the else branch %s." new_ty_then.id new_ty_else.id) main_ext;
            Let(new_pat,new_t,new_t_then, Some new_t_else), new_ty_then
      end
  | PPLetFilter(var_list,cond,t_then,t_else_op) ->
      let new_var_list,new_env =
        List.fold_left (fun (acc_vars,acc_env) (id,ty) ->
          let new_ty = get_type_from_id env ty in
          let new_id = fresh_id_from_ident id in
          let new_var = { id_v = new_id; typ_v = new_ty; status_v = SCNone; options_v = []; side_v = NoRelevant } in
          (acc_vars @ [new_var], StringMap.add new_id.id (EVar new_var) acc_env)
        ) ([],env) var_list
      in
      let new_cond, cond_type = expand_pterm_e new_env cond in

      if not (is_equal_type cond_type bool_id)
      then input_error (Printf.sprintf "The predicate of the Let is expected to be of type bool but is of type %s" cond_type.id) main_ext;

      let new_t_then, then_type = expand_pterm_e new_env t_then in
      let new_t_else_op = match t_else_op with
        | None -> None
        | Some t_else ->
            let new_t_else, else_type = expand_pterm_e env t_else in
            if not (is_equal_type then_type else_type)
            then input_error (Printf.sprintf "The type of the then branch (%s) differs from the type of the else branch (%s)." then_type.id else_type.id) main_ext;
            Some new_t_else
      in

      LetFilter(new_var_list,new_cond,new_t_then,new_t_else_op), then_type

and expand_pterm_e env (pterm,ext) = expand_pterm env ext pterm

and expand_tpattern env t_match_op = function
  | PPatVar((str,ext),ty_op) ->
      let new_type =
        match t_match_op,ty_op with
          | None, None -> input_error (Printf.sprintf "The variable %s does not have a type." str) ext
          | Some t_match, None -> get_type t_match
          | Some t_match, Some ty ->
              let t_type = get_type t_match in
              let new_ty_op = get_type_from_id env ty in
              if not (is_equal_type new_ty_op t_type)
              then input_error (Printf.sprintf "The variable %s is declared with type %s but is assigned a term of type %s" str new_ty_op.id t_type.id) ext;
              t_type
          | None, Some ty -> get_type_from_id env ty
      in
      let new_var_id = fresh_id_from_ident (str,ext) in
      let new_var = { id_v = new_var_id; typ_v = new_type; status_v = SCNone; options_v = []; side_v = NoRelevant } in
      let new_env = StringMap.add str (EVar new_var) env in
      (PatVar new_var, new_env)
  | PPatTuple(pat_args) ->
      begin match t_match_op with
        | Some (Tuple (args)) when List.length pat_args = List.length args ->
            let (new_pat_args,new_env) =
              List.fold_left2 (fun (acc_pat,acc_env) pat t ->
                let new_pat,new_env = expand_tpattern acc_env (Some t) pat in
                (acc_pat @ [new_pat],new_env)
              ) ([],env) pat_args args
            in
            PatTuple(new_pat_args), new_env
        | _ ->
          let (new_pat_args,new_env) =
            List.fold_left (fun (acc_pat,acc_env) pat ->
              let new_pat,new_env = expand_tpattern acc_env None pat in
              (acc_pat @ [new_pat],new_env)
            ) ([],env) pat_args
          in
          PatTuple(new_pat_args), new_env
      end
  | PPatChoiceApp(pat1,pat2) ->
    begin
      match pat1, pat2 with
      | PPatVar((_,_),_), PPatVar((_,_),_) ->
        let pat1',env1' = expand_tpattern env t_match_op pat1 in
        let pat2',env2' = expand_tpattern env1' t_match_op pat2 in
        PatChoiceApp(pat1',pat2'), env2'
      | _ -> failwith "[transform.ml >> expand_tpattern] Syntax error in input"
    end
  | PPatEqual term ->
      let (new_term,_) = expand_pterm_e env term in
      PatEq new_term, env

let rec expand_tprocess env = function
  | PNil -> { proc = CNil; comment = None }
  | PPar(p1,p2) -> { proc = CPar (expand_tprocess env p1,expand_tprocess env p2); comment = None }
  | PRepl(p) -> { proc = CRepl (expand_tprocess env p); comment = None }
  | PRestr((str,ext_n),vars,id_ty,op_l,p) ->
      let new_id = fresh_id_from_ident (str,ext_n) in
      let new_vars =
        match vars with
          | None -> None
          | Some vars_l -> Some (List.map (get_variable_from_id env) vars_l)
      in
      let new_type = get_type_from_id env id_ty in

      let new_op_l =
        List.map (fun (str,ext) ->
          match str with
            | _ -> input_error "Unexpected option for a name." ext
        ) op_l
      in

      let status_channel = SCNone
      in

      let new_name = { id_n = new_id; typ_n = new_type; options_n = new_op_l; status_n = status_channel; ext_n = ext_n } in
      let new_env = StringMap.add str (EName new_name) env in
      let new_p = expand_tprocess new_env p in

      { proc = CRestr(new_name,new_vars,new_p); comment = None }
  | PLetDef((str,ext),args) ->
      let (_,ty_l,gen_proc) = get_process_from_id env (str,ext) in

      if List.length ty_l <> List.length args
      then input_error (Printf.sprintf "The function %s is expecting %d arguments but is given %d." str (List.length ty_l) (List.length args)) ext;

      let new_args =
        List.fold_left2 (fun acc_arg (t,ext_t) ty ->
          let new_t, new_type = expand_pterm_e env (t,ext_t) in
          if not (is_equal_type new_type ty)
          then input_error (Printf.sprintf "The argument is expected to be of type %s but is of type %s" ty.id new_type.id) ext_t;

          acc_arg @ [ new_t ]
        ) [] args ty_l
      in

      let new_proc = gen_proc new_args in
      { new_proc with comment = Some (Printf.sprintf "Application of the process %s" str) }
  | PTest((cond,ext_cond),p_then,p_else) ->
      let new_cond, cond_type = expand_pterm_e env (cond,ext_cond) in
      if not (is_equal_type cond_type bool_id)
      then input_error (Printf.sprintf "The condition of the test is expected to be of type bool but is of type %s" cond_type.id) ext_cond;

      { proc = CTest(new_cond,expand_tprocess env p_then,expand_tprocess env p_else); comment = None }
  | PInput((c,ext_c),pat,p) ->
      let new_c, c_type = expand_pterm_e env (c,ext_c) in
      if not (is_equal_type c_type channel_id)
      then input_error (Printf.sprintf "The term is expected to be of type channel but is of type %s" c_type.id) ext_c;
      let new_pat,new_env = expand_tpattern env None pat in
      let new_p = expand_tprocess new_env p in

      { proc = CInput(new_c,new_pat,new_p); comment = None }
  | POutput((c,ext_c),t,p) ->
      let new_c, c_type = expand_pterm_e env (c,ext_c) in
      if not (is_equal_type c_type channel_id)
      then input_error (Printf.sprintf "The term is expected to be of type channel but is of type %s" c_type.id) ext_c;
      let new_t,_ = expand_pterm_e env t in
      let new_p = expand_tprocess env p in

      { proc = COutput(new_c,new_t,new_p); comment = None }
  | PLet(pat,t,p_then,p_else) ->
      let new_t,_ = expand_pterm_e env t in
      let new_pat, new_env = expand_tpattern env (Some new_t) pat in
      let new_p_then = expand_tprocess new_env p_then in
      let new_p_else = expand_tprocess env p_else in

      { proc = CLet(new_pat,new_t,new_p_then,new_p_else); comment = None }
  | PLetFilter(var_list,(cond,ext_cond),p_then,p_else) ->
      let new_var_list,new_env =
        List.fold_left (fun (acc_vars,acc_env) (id,ty) ->
          let new_ty = get_type_from_id env ty in
          let new_id = fresh_id_from_ident id in
          let new_var = { id_v = new_id; typ_v = new_ty; status_v = SCNone; options_v = []; side_v = NoRelevant } in
          (acc_vars @ [new_var], StringMap.add new_id.id (EVar new_var) acc_env)
        ) ([],env) var_list
      in

      let new_cond, cond_type = expand_pterm_e new_env (cond,ext_cond) in

      if not (is_equal_type cond_type bool_id)
      then input_error (Printf.sprintf "The predicate is expected to be of type bool but is of type %s" cond_type.id) ext_cond;

      let new_p_then = expand_tprocess new_env p_then in
      let new_p_else = expand_tprocess env p_else in

      { proc = CLetFilter(new_var_list,new_cond,new_p_then,new_p_else); comment = None }
  | PEvent((str,ext),args,vars_op,p) ->
      let ev = get_event_from_id env (str,ext) in

      if List.length ev.arg_ty_e <> List.length args
      then input_error (Printf.sprintf "The event %s is expecting %d arguments but is given %d." str (List.length ev.arg_ty_e) (List.length args)) ext;

      let new_args =
        List.fold_left2 (fun acc_arg (t,ext_t) ty ->
          let new_t, new_type = expand_pterm_e env (t,ext_t) in
          if not (is_equal_type new_type ty)
          then input_error (Printf.sprintf "The argument is expected to be of type %s but is of type %s" ty.id new_type.id) ext_t;

          acc_arg @ [ new_t ]
        ) [] args ev.arg_ty_e
      in

      let new_vars_op = match vars_op with
        | None -> None
        | Some vars -> Some (List.map (get_variable_from_id env) vars)
      in
      let new_p = expand_tprocess env p in

      { proc = CEvent(ev,new_args,new_vars_op,new_p); comment = None }
  | PPhase(i,p) -> { proc = CPhase(i, expand_tprocess env p); comment = None }
  | PBarrier _ -> input_error "The front-end does not handle synchronisation yet." dummy_ext
  | PInsert ((str,ext),args,p) ->
      let tbl = get_table_from_id env (str,ext) in

      if List.length tbl.arg_ty_t <> List.length args
      then input_error (Printf.sprintf "The table %s is expecting %d arguments but is given %d." str (List.length tbl.arg_ty_t) (List.length args)) ext;

      let new_args =
        List.fold_left2 (fun acc_arg (t,ext_t) ty ->
          let new_t, new_type = expand_pterm_e env (t,ext_t) in
          if not (is_equal_type new_type ty)
          then input_error (Printf.sprintf "The argument is expected to be of type %s but is of type %s" ty.id new_type.id) ext_t;

          acc_arg @ [ new_t ]
        ) [] args tbl.arg_ty_t
      in

      let new_p = expand_tprocess env p in

      { proc = CInsert(tbl,new_args,new_p); comment = None }
  | PGet((str,ext),pat_l,cond,p_then,p_else) ->
      let tbl = get_table_from_id env (str,ext) in

      if List.length tbl.arg_ty_t <> List.length pat_l
      then input_error (Printf.sprintf "The table %s is expecting %d arguments but is given %d." str (List.length tbl.arg_ty_t) (List.length pat_l)) ext;

      let new_pat_l, new_env =
        List.fold_left2 (fun (acc_pat,acc_env) pat ty ->
          let v = fresh_var ty NoRelevant in
          let new_pat,new_env = expand_tpattern acc_env (Some (Var v)) pat in
          (acc_pat @ [new_pat], new_env)
        ) ([],env) pat_l tbl.arg_ty_t
      in

      let new_cond = match cond with
        | None -> None
        | Some ((cond,ext_cond)) ->
            let new_cond,cond_type = expand_pterm_e new_env (cond,ext_cond) in
            if not (is_equal_type cond_type bool_id)
            then input_error (Printf.sprintf "The predicate is expected to be of type bool but is of type %s" cond_type.id) ext_cond;
            Some new_cond
      in

      let new_p_then = expand_tprocess new_env p_then in
      let new_p_else = expand_tprocess env p_else in

      { proc = CGet(tbl,new_pat_l,new_cond,new_p_then,new_p_else); comment = None }



let rec gather_restriction env = function
  | PNil -> ()
  | PPar(p1,p2)
  | PTest(_,p1,p2)
  | PLet(_,_,p1,p2)
  | PLetFilter(_,_,p1,p2)
  | PGet(_,_,_,p1,p2) -> gather_restriction env p1; gather_restriction env p2
  | PRepl p
  | PInput(_,_,p)
  | POutput(_,_,p)
  | PEvent(_,_,_,p)
  | PInsert(_,_,p) -> gather_restriction env p
  | PPhase _ -> input_error "The front-end does not handle phases yet." dummy_ext
  | PBarrier _ -> input_error "The front-end does not handle synchronisation yet." dummy_ext
  | PRestr((str,_),_,id_ty,_,p) ->
      let new_type = get_type_from_id env id_ty in

      restricted_name := StringMap.add str new_type !restricted_name;
      gather_restriction env p;
  | PLetDef _ -> ( )


let rec expand_tdecl env proc = function
  | [] -> [DProcess (expand_tprocess env proc)]
  | decl :: q_decl ->
      match decl with
        | TTypeDecl id ->
            let str,_ = id in
            let new_id = fresh_id_from_ident id in
            (DType new_id)::expand_tdecl (StringMap.add str (EType new_id) env) proc q_decl
        | TFunDecl(id,typ_args,typ,options) ->
            let (str,_) = id in
            let typ_args_id = List.map (get_type_from_id env) typ_args in
            let typ_id = get_type_from_id env typ in
            let is_data = ref false
            and is_priv = ref false
            and is_type_conv = ref false in
            List.iter (function
              | "private",_ -> is_priv := true
              | "data",_ -> is_data := true
              | "typeConverter",_ -> is_type_conv := true
              | s,ext -> input_error (Printf.sprintf "A function cannot be declared with the option %s." s) ext
            ) options;
            let f =
              { id_f = fresh_id_from_ident id;
                arg_ty = typ_args_id;
                return_ty = typ_id;
                priv = !is_priv;
                pure_priv = !is_priv;
                data = !is_data;
                type_conv = !is_type_conv;
                funcat = Const
              }
            in
            (DFun f)::(expand_tdecl (StringMap.add str (EFun f) env) proc q_decl)
        | TEventDecl(id,typ_args) ->
            let (str,_) = id in
            let typ_args_id = List.map (get_type_from_id env) typ_args in
            let new_id = fresh_id_from_ident id in
            let new_ev = { id_e = new_id; arg_ty_e = typ_args_id; side_e = NoRelevant } in

            (DEvent new_ev)::(expand_tdecl (StringMap.add str (EEvent new_ev) env) proc q_decl)
        | TConstDecl(id,typ,options) ->
            let str,_ = id in
            let typ_id = get_type_from_id env typ in
            let is_data = ref false
            and is_priv = ref false
            and is_type_conv = ref false in
            List.iter (function
              | "private",_ -> is_priv := true
              | "data",_ -> is_data := true
              | "typeConverter",_ -> is_type_conv := true
              | s,ext -> input_error (Printf.sprintf "A constant cannot be declared with the option %s." s) ext
            ) options;
            let f =
              { id_f = fresh_id_from_ident id;
                arg_ty = [];
                return_ty = typ_id;
                pure_priv = !is_priv;
                priv = !is_priv;
                data = !is_data;
                type_conv = !is_type_conv;
                funcat = Const
              }
            in
            (DConst f)::(expand_tdecl (StringMap.add str (EFun f) env) proc q_decl)
        | TReduc(reduc,options) ->
            let typ_lhs = ref [] in
            let typ_rhs = ref None in
            let f_str = ref None in
            let symb = ref None in

            let new_reduc =
              List.map (fun (vars,(lhs,ext),rhs) ->
                let new_env, new_vars =
                  List.fold_left (fun (acc_env,acc_var) ((str,ext),typ) ->
                    let new_id = fresh_id_from_ident (str,ext) in
                    let typ_id = get_type_from_id acc_env typ in
                    let v = { id_v = new_id; typ_v = typ_id; status_v = SCNone; options_v = []; side_v = NoRelevant } in
                    (StringMap.add str (EVar v) acc_env, acc_var @ [v])
                  ) (env,[]) vars
                in
                match lhs with
                  | PFunApp((str,f_ext),args) ->
                      begin match !f_str with
                        | None -> f_str := Some str
                        | Some str' when str = str' -> ()
                        | _ -> input_error "The head symbol is not the same in the declaration." f_ext
                      end;
                      let (new_args,typ_args) =
                        List.fold_left (fun (acc_args,acc_typ) t ->
                          let (new_t,new_typ) = expand_and_check_pure_priv_term_e new_env t in
                          (acc_args @ [new_t], acc_typ @ [new_typ])
                        ) ([],[]) args
                      and (new_rhs,new_typ_rhs) = expand_and_check_pure_priv_term_e new_env rhs in

                      if !typ_lhs = [] then typ_lhs := typ_args;
                      if not (is_equal_type_list !typ_lhs typ_args) then input_error "The arguments of the left hand sides of the reduction do not have the same type." f_ext;

                      begin match !typ_rhs with
                        | None -> typ_rhs := Some new_typ_rhs
                        | Some typ when is_equal_type typ new_typ_rhs -> ()
                        | _ -> input_error "The right hand sides of the reduction do not have the same type." f_ext
                      end;

                      let f = match !symb with
                        | None ->
                            let f =
                              { id_f = fresh_id_from_ident (str,f_ext);
                                arg_ty = !typ_lhs;
                                return_ty = (match !typ_rhs with Some t -> t | None -> internal_error "[transform.ml >> expand_tdecl] Unexpected case 1");
                                priv = (match options with ["private",_] -> true | _ -> false);
                                pure_priv = false;
                                data = false;
                                type_conv = false;
                                funcat = Dest
                              }
                            in
                            symb := Some f;
                            f
                        | Some f ->
                          f.funcat <- Dest;
                          f
                      in
                      (new_vars,Fun(f,new_args),new_rhs)
                  | _ -> input_error "The term should be an application of a function." ext
              ) reduc
            in

            begin match !symb with
              | None -> internal_error "[transform.ml >> expand_tdecl] Unexpected case 2"
              | Some f -> (DReduc (f,new_reduc))::(expand_tdecl (StringMap.add f.id_f.id (EFun f) env) proc q_decl)
            end
        | TReducFail(f,typ_args,typ_r,reduc,options) ->
            let new_f = fresh_id_from_ident f in
            let new_typ_args = List.map (get_type_from_id env) typ_args in
            let new_typ_r = get_type_from_id env typ_r in

            let f =
              {
                id_f = new_f;
                arg_ty = new_typ_args;
                return_ty = new_typ_r;

                priv = (match options with ["private",_] -> true | _ -> false);
                data = false;
                type_conv = false;
                pure_priv = false;
                funcat = Dest
              }
            in

            let new_reduc =
              List.map (fun (may_vars,(lhs,ext),rhs) ->
                let new_env, new_may_vars =
                  List.fold_left (fun (acc_env,acc_may_var) ((str,ext),typ,b) ->
                    let new_id = fresh_id_from_ident (str,ext) in
                    let typ_id = get_type_from_id acc_env typ in
                    let v = { id_v = new_id; typ_v = typ_id; status_v = SCNone; options_v = []; side_v = NoRelevant } in
                    (StringMap.add str (EVar v) acc_env, acc_may_var @ [v,b])
                  ) (env,[]) may_vars
                in
                match lhs with
                  | PFunApp((str,f_ext),args) ->
                      if f.id_f.id <> str then input_error "The head symbol is not the same in the declaration." f_ext;

                      let (new_args,new_typ_args') =
                        List.fold_left (fun (acc_args,acc_typ) t ->
                          let (new_t,new_typ) = expand_and_check_pure_priv_term_e new_env t in
                          (acc_args @ [new_t], acc_typ @ [new_typ])
                        ) ([],[]) args
                      and (new_rhs,new_typ_rhs) = expand_and_check_pure_priv_term_e new_env rhs in

                      if not (is_equal_type_list new_typ_args new_typ_args') then input_error "The arguments of the left hand sides of the reduction do not have the same type." f_ext;
                      if not (is_equal_type new_typ_r new_typ_rhs) then input_error "The right hand sides of the reduction do not have the same type." f_ext;

                      (new_may_vars,Fun(f,new_args),new_rhs)
                  | _ -> input_error "The term should be an application of a function." ext
              ) reduc
            in

            (DReducFail(f,new_reduc))::(expand_tdecl (StringMap.add new_f.id (EFun f) env) proc q_decl)
        | TEquation(eq_list,options) ->
            let new_eq_list =
              List.map (fun (vars,lhs,rhs) ->
                let new_env, new_vars =
                  List.fold_left (fun (acc_env,acc_var) ((str,ext),typ) ->
                    let new_id = fresh_id_from_ident (str,ext) in
                    let typ_id = get_type_from_id acc_env typ in
                    let v = { id_v = new_id; typ_v = typ_id; status_v = SCNone; options_v = []; side_v = NoRelevant } in
                    (StringMap.add str (EVar v) acc_env, acc_var @ [v])
                  ) (env,[]) vars
                in

                let new_lhs,_ = expand_and_check_pure_priv_term_e new_env lhs in
                let new_rhs,_ = expand_and_check_pure_priv_term_e new_env rhs in
                (new_vars,new_lhs,new_rhs)
              ) eq_list
            in

            (DEquation(new_eq_list,options))::(expand_tdecl env proc q_decl)
        | TPredDecl(id,typ_args,options) ->
            let new_id = fresh_id_from_ident id in
            let new_typ_args = List.map (get_type_from_id env) typ_args in
            let f =
              {
                id_f = new_id;
                arg_ty = new_typ_args;
                return_ty = bool_id;
                pure_priv = false;
                priv = false;
                data = false;
                type_conv = false;
                funcat = Const
              }
            in

            (DPred(f,options))::(expand_tdecl (StringMap.add new_id.id (EFun f) env) proc q_decl)
        | TTableDecl(id,typ_args) ->
            let new_id = fresh_id_from_ident id in
            let new_typ_args = List.map (get_type_from_id env) typ_args in
            let precise_options = ref TblNone in
            let uaction = ref false in

            let status = match !precise_options with
              | TblNone -> STNone
            in

            let tbl =
              {
                id_t = new_id;
                arg_ty_t = new_typ_args;
                options_t = !precise_options;
                uaction_t = !uaction;
                status_t = status
              }
            in

            (DTable tbl)::(expand_tdecl (StringMap.add new_id.id (ETable tbl) env) proc q_decl)
        | TSet (id,value) -> (DSet (id,value))::(expand_tdecl env proc q_decl)
        | TPDef(id,may_vars,p) ->
            let new_id = fresh_id_from_ident id in

            let new_typ_args = List.map (fun (_,typ,_) -> get_type_from_id env typ) may_vars in

            let generate_process args_list =
              let rec gen_env env typ_l vars args = match typ_l,vars,args with
                | [],[],[] -> env
                | [],_,_ | _,[],_ | _,_,[] -> internal_error "[transform.ml >> expand_tdecl] Uncorrect list sizes."
                | ty::q_t, ((str_v,_),_,_)::q_v, arg::q_a -> gen_env (StringMap.add str_v (EArg (arg,ty)) env) q_t q_v q_a
              in
              let new_env = gen_env env new_typ_args may_vars args_list in
              expand_tprocess new_env p
            in

            gather_restriction env p;

            expand_tdecl (StringMap.add new_id.id (EProcess(new_id,new_typ_args,generate_process)) env) proc q_decl
        | TQuery _
        | TNoninterf _
        | TWeaksecret _
        | TNoUnif _
        | TDefine _
        | TClauses _
        | TExpand _ -> internal_error "[transform.ml >> expand_tdecl] Wrong declaration"
        | TElimtrue(may_vars,t) ->
            let new_env, new_may_vars =
              List.fold_left (fun (acc_env,acc_may_var) ((str,ext),typ,b) ->
                let new_id = fresh_id_from_ident (str,ext) in
                let typ_id = get_type_from_id acc_env typ in
                let v = { id_v = new_id; typ_v = typ_id; status_v = SCNone; options_v = []; side_v = NoRelevant } in
                (StringMap.add str (EVar v) acc_env, acc_may_var @ [v,b])
              ) (env,[]) may_vars
            in
            let t',_ = expand_term_e new_env t in
            (DElimtrue(new_may_vars,t'))::(expand_tdecl env proc q_decl)
        | TFree((str,ext),typ,options) ->
            let new_id = fresh_id_from_ident (str,ext) in
            let new_typ = get_type_from_id env typ in
            let is_private = ref false in
            let new_options =
              List.map (function
                | "private",_ -> is_private := true; NDPrivate
                | _,ext ->
                  input_error
                    (Printf.sprintf "Wrong option for a name: A name can have the option private."
                    ) ext
              ) options
            in

            let status_n = SCNone
            in

            let new_n =
              {
                id_n = new_id;
                typ_n = new_typ;
                options_n = new_options;
                status_n = status_n;
                ext_n = ext
              }
            in

            (DFree new_n)::(expand_tdecl (StringMap.add new_id.id (EName new_n) env) proc q_decl)
        | TLetFun(id,may_vars,t) ->
            let new_id = fresh_id_from_ident id in
            let new_env, new_may_vars =
              List.fold_left (fun (acc_env,acc_may_var) ((str,ext),typ,b) ->
                let new_id = fresh_id_from_ident (str,ext) in
                let typ_id = get_type_from_id acc_env typ in
                let v = { id_v = new_id; typ_v = typ_id; status_v = SCNone; options_v = []; side_v = NoRelevant } in
                (StringMap.add str (EVar v) acc_env, acc_may_var @ [v,b])
              ) (env,[]) may_vars
            in

            let new_typ_args = List.map (fun (v,_) -> v.typ_v) new_may_vars in

            let new_term,return_typ = expand_pterm_e new_env t in

            let f =
              {
                id_f = new_id;
                arg_ty = new_typ_args;
                return_ty = return_typ;
                priv = false;
                data = false;
                type_conv = false;
                pure_priv = false;
                funcat = Const
              }
            in

            let new_env' = StringMap.add new_id.id (EFun f) env in

            (DLetFun (f,new_may_vars,new_term))::(expand_tdecl new_env' proc q_decl)
        | _ -> failwith "[transform.ml >> expand_tdecl] Syntax not supported"

let rec set_ignore_type = function
  | [] -> ()
  | DSet(id,value)::q ->
      begin match id,value with
        | ("ignoreTypes",_), S ("true",_)
        | ("ignoreTypes",_), S ("all",_) -> ignore_type := true
        | ("ignoreTypes",_), S ("false",_)
        | ("ignoreTypes",_), S ("none",_)
        | ("ignoreTypes",_), S ("attacker",_) -> ignore_type := false
        | _,_ -> ()
      end;
      set_ignore_type q
  | _::q -> set_ignore_type q

let set_allowDiffPatterns decl =
  if !allowDiffPatterns then
    DSet(("allowDiffPatterns",dummy_ext),S ("true",dummy_ext))::decl
  else decl

let set_reconstructTrace decl =
  if not(!reconstructTrace) then
    DSet(("reconstructTrace",dummy_ext),S ("false",dummy_ext))::decl
  else decl



let rec get_vars_term = function
| Var x -> [x]
| DumVar x -> [x]
| Name _ -> []
| Sum(_,_) -> []
| Nat _ -> []
| Fun(_,args) -> List.fold_left (fun acc t -> List.append acc (get_vars_term t)) [] args
| Tuple(args) -> List.fold_left (fun acc t -> List.append acc (get_vars_term t)) [] args
| Restr(_,_,_) -> []
| Test(cond,t_then,t_else_op) -> List.append (get_vars_term cond) (List.append (get_vars_term t_then) (match t_else_op with | None -> [] | Some t_else -> get_vars_term t_else))
| Let(pat,cond,t_then,t_else_op) -> List.append (get_vars_pattern pat) (List.append (get_vars_term cond) (List.append (get_vars_term t_then) (match t_else_op with | None -> [] | Some t_else -> get_vars_term t_else)))
| _ -> failwith "[transform.ml >> get_vars_term] Syntax not supported"

and get_vars_pattern = function
  | PatVar x -> [x]
  | PatEq _ -> []
  | PatTuple l_pat -> List.fold_left (fun acc pat -> List.append acc (get_vars_pattern pat)) [] l_pat
  | PatChoiceApp (pat1, pat2) ->
    let acc1 = get_vars_pattern pat1 in
    let acc2 = get_vars_pattern pat2 in
    List.append acc1 acc2
  | PatData _ -> failwith "[transform.ml >> get_vars_pattern] Syntax not supported (Data)"

and get_vars_names_process = function
  | CNil -> [],[]
  | CPar (p1,p2) ->
    let vars1,names1 = get_vars_names_process p1.proc in
    let vars2,names2 = get_vars_names_process p2.proc in
    List.append vars1 vars2, List.append names1 names2
  | CRepl p -> get_vars_names_process p.proc
  | CRestr (n, _, p) ->
    let vars,names = get_vars_names_process p.proc in
    vars, n::names
  | CTest (_, p_then, p_else) ->
    let vars1,names1 = get_vars_names_process p_then.proc in
    let vars2,names2 = get_vars_names_process p_else.proc in
    List.append vars1 vars2, List.append names1 names2
  | CLet (pat,_,p_then,p_else) ->
    let vars = get_vars_pattern pat in
    let vars1,names1 = get_vars_names_process p_then.proc in
    let vars2,names2 = get_vars_names_process p_else.proc in
    List.append (List.append vars vars1) vars2, List.append names1 names2
  | CInsert(_,_,p) -> get_vars_names_process p.proc
  | CInput(_,pat,p) ->
    let vars = get_vars_pattern pat in
    let vars1,names1 = get_vars_names_process p.proc in
    List.append vars vars1, names1
  | COutput(_,_,p) -> get_vars_names_process p.proc
  | CEvent(_,_,_,p) -> get_vars_names_process p.proc
  | CGet(_, pat_list, _, p_then, p_else) ->
    let vars = List.fold_left (fun acc pat -> List.append acc (get_vars_pattern pat)) [] pat_list in
    let vars1,names1 = get_vars_names_process p_then.proc in
    let vars2,names2 = get_vars_names_process p_else.proc in
    List.append (List.append vars vars1) vars2, List.append names1 names2
  | CPhase(_,p) -> get_vars_names_process p.proc

  | CLetFilter _ -> failwith "[transform.ml >> get_vars_names_process] Let Filter not supported"


let rec refresh_vars_names_pattern side vars_subst names_subst = function
  | PatVar x ->
    begin
      try
        let _, new_xL, new_xR = List.find (fun (v,_,_) -> v == x) vars_subst in
        begin match new_xR with | None -> PatVar new_xL | Some xR -> PatChoiceApp (PatVar new_xL, PatVar xR)
        end
      with
      | Not_found -> PatVar x
    end
  | PatEq t -> PatEq (refresh_vars_names_terms side vars_subst names_subst t)
  | PatTuple l_pat -> PatTuple (List.map (fun pat -> refresh_vars_names_pattern side vars_subst names_subst pat) l_pat)
  | PatChoiceApp (pat1, pat2) -> PatChoiceApp (refresh_vars_names_pattern side vars_subst names_subst pat1, refresh_vars_names_pattern side vars_subst names_subst pat2)
  | PatData _ -> failwith "[transform.ml >> refresh_vars_names_pattern] Syntax not supported (Data)"

and refresh_vars_names_terms side vars_subst names_subst = function
  | Var x ->
    begin
      try
        let _, new_xL, new_xR = List.find (fun (v,_,_) -> v == x) vars_subst in
        begin match new_xR with | None -> Var new_xL | Some xR -> Fun(choice_symb,[Var new_xL; Var xR])
        end
      with
      | Not_found -> Var x
    end
  | DumVar x ->
    begin
      try
        let _, new_xL, new_xR = List.find (fun (v,_,_) -> v == x) vars_subst in
        begin match new_xR with | None -> DumVar new_xL | Some xR -> Fun(choice_symb,[DumVar new_xL; DumVar xR])
        end
      with
      | Not_found -> DumVar x
    end
  | Name n ->
    begin
      try
        let _, new_n = List.find (fun (m,_) -> m == n) names_subst in
        Name new_n
      with
      | Not_found -> Name n
    end
  | Sum(v,n) -> Sum(v,n)
  | Nat n -> Nat n
  | Fun(f,[t1;t2]) when List.exists (fun f' -> f == f') [choice_symb] ->
    begin match side with
    | NoRelevant -> Fun(f, [refresh_vars_names_terms side vars_subst names_subst t1; refresh_vars_names_terms side vars_subst names_subst t2])
    | Left -> Fun(f, [refresh_vars_names_terms side vars_subst names_subst t1; t2])
    | Right -> Fun(f, [t1; refresh_vars_names_terms side vars_subst names_subst t2])
    end
  | Fun(f,args) ->
    Fun(f, List.map (fun t -> refresh_vars_names_terms side vars_subst names_subst t) args)
  | Tuple(args) -> Tuple(List.map (fun t -> refresh_vars_names_terms side vars_subst names_subst t) args)
  | Restr(n,op,t) ->
    begin
      try
        let _, new_n = List.find (fun (m,_) -> m == n) names_subst in
        let new_op = match op with
          | None -> None
          | Some l -> Some (List.concat (List.map (fun x -> begin
            try
              let _, new_xL, new_xR = List.find (fun (v,_,_) -> v == x) vars_subst in
              begin match new_xR with | None -> [new_xL] | Some xR -> [new_xL;xR]
              end
            with
            | Not_found -> [x]
          end) l))
        in
        Restr(new_n,new_op,refresh_vars_names_terms side vars_subst names_subst t)
      with
      | Not_found ->
        let new_op = match op with
          | None -> None
          | Some l -> Some (List.concat (List.map (fun x -> begin
            try
              let _, new_xL, new_xR = List.find (fun (v,_,_) -> v == x) vars_subst in
              begin match new_xR with | None -> [new_xL] | Some xR -> [new_xL;xR]
              end
            with
            | Not_found -> [x]
          end) l))
        in
        Restr(n,new_op,refresh_vars_names_terms side vars_subst names_subst t)
    end
  | Test(cond,t_then,t_else_op) ->
      begin
        match t_else_op with
        | None -> Test(refresh_vars_names_terms side vars_subst names_subst cond, refresh_vars_names_terms side vars_subst names_subst t_then, None)
        | Some t_else -> Test(refresh_vars_names_terms side vars_subst names_subst cond, refresh_vars_names_terms side vars_subst names_subst t_then, Some (refresh_vars_names_terms side vars_subst names_subst t_else))
      end
  | Let(pat,cond,t_then,t_else_op) ->
    begin
      match t_else_op with
      | None -> Let(refresh_vars_names_pattern side vars_subst names_subst pat, refresh_vars_names_terms side vars_subst names_subst cond, refresh_vars_names_terms side vars_subst names_subst t_then, None)
      | Some t_else -> Let(refresh_vars_names_pattern side vars_subst names_subst pat, refresh_vars_names_terms side vars_subst names_subst cond, refresh_vars_names_terms side vars_subst names_subst t_then, Some (refresh_vars_names_terms side vars_subst names_subst t_else))
    end
  | _ -> failwith "[transform.ml >> refresh_vars_names_terms] Syntax not supported"

let rec refresh_vars_names_proc_rec side vars_subst names_subst = function
  | CNil -> CNil
  | CPar (p1,p2) -> CPar(refresh_vars_names_proc_rec_c side vars_subst names_subst p1, refresh_vars_names_proc_rec_c side vars_subst names_subst p2)
  | CRepl p -> CRepl (refresh_vars_names_proc_rec_c side vars_subst names_subst p)
  | CRestr (n, vars, p) ->
    begin
      try
        let _, new_n = List.find (fun (m,_) -> m == n) names_subst in
        let new_op = match vars with
          | None -> None
          | Some l -> Some (List.concat (List.map (fun x -> begin
            try
              let _, new_xL, new_xR = List.find (fun (v,_,_) -> v == x) vars_subst in
              begin match new_xR with | None -> [new_xL] | Some xR -> [new_xL;xR]
              end
            with
            | Not_found -> [x]
          end) l))
        in
        CRestr(new_n,new_op,refresh_vars_names_proc_rec_c side vars_subst names_subst p)
      with
      | Not_found ->
        let new_op = match vars with
          | None -> None
          | Some l -> Some (List.concat (List.map (fun x -> begin
            try
              let _, new_xL, new_xR = List.find (fun (v,_,_) -> v == x) vars_subst in
              begin match new_xR with | None -> [new_xL] | Some xR -> [new_xL;xR]
              end
            with
            | Not_found -> [x]
          end) l))
        in
        CRestr(n,new_op,refresh_vars_names_proc_rec_c side vars_subst names_subst p)
    end
  | CTest (t, p_then, p_else) ->
    CTest (
    refresh_vars_names_terms side vars_subst names_subst t,
    refresh_vars_names_proc_rec_c side vars_subst names_subst p_then,
    refresh_vars_names_proc_rec_c side vars_subst names_subst p_else
    )
  | CLet (pat,t,p_then,p_else) ->
    CLet (
    refresh_vars_names_pattern side vars_subst names_subst pat,
    refresh_vars_names_terms side vars_subst names_subst t,
    refresh_vars_names_proc_rec_c side vars_subst names_subst p_then,
    refresh_vars_names_proc_rec_c side vars_subst names_subst p_else
    )
  | CInsert(tab,l,p) ->
    CInsert(
    tab,
    List.map (fun t -> refresh_vars_names_terms side vars_subst names_subst t) l,
    refresh_vars_names_proc_rec_c side vars_subst names_subst p
    )
  | CInput(c,pat,p) ->
    CInput(
    refresh_vars_names_terms side vars_subst names_subst c,
    refresh_vars_names_pattern side vars_subst names_subst pat,
    refresh_vars_names_proc_rec_c side vars_subst names_subst p
    )
  | COutput(c,t,p) ->
    COutput(
    refresh_vars_names_terms side vars_subst names_subst c,
    refresh_vars_names_terms side vars_subst names_subst t,
    refresh_vars_names_proc_rec_c side vars_subst names_subst p
    )
  | CEvent(ev,args,None,p) ->
    CEvent(
    ev,
    List.map (fun t -> refresh_vars_names_terms side vars_subst names_subst t) args,
    None,
    refresh_vars_names_proc_rec_c side vars_subst names_subst p
    )
  | CEvent(_,_,Some _,_) -> failwith "[transform.ml >> refresh_vars_names_proc_rec] Injective events not supported"
  | CGet(tab, pat_list, None, p_then, p_else) ->
    CGet(
    tab,
    List.map (fun pat -> refresh_vars_names_pattern side vars_subst names_subst pat) pat_list,
    None,
    refresh_vars_names_proc_rec_c side vars_subst names_subst p_then,
    refresh_vars_names_proc_rec_c side vars_subst names_subst p_else
    )
  | CGet(tab, pat_list, Some phi, p_then, p_else) ->
    CGet(
    tab,
    List.map (fun pat -> refresh_vars_names_pattern side vars_subst names_subst pat) pat_list,
    Some (refresh_vars_names_terms side vars_subst names_subst phi),
    refresh_vars_names_proc_rec_c side vars_subst names_subst p_then,
    refresh_vars_names_proc_rec_c side vars_subst names_subst p_else
    )
  | CPhase(i,p) -> CPhase(i,refresh_vars_names_proc_rec_c side vars_subst names_subst p)
  | CLetFilter _ -> failwith "[transform.ml >> refresh_vars_names_proc_rec] Let Filter not supported"


and refresh_vars_names_proc_rec_c side vars_subst names_subst proc = { proc = refresh_vars_names_proc_rec side vars_subst names_subst proc.proc; comment=None }


let refresh_vars_proc_choice proc =
  let vars,_ = get_vars_names_process proc.proc in
  let new_vars_subst = List.map (fun x -> x,fresh_var x.typ_v Left, Some (fresh_var x.typ_v Right)) vars in
  (*let new_names_subst = List.map (fun n -> n,fresh_name n.typ_n) names in*)
  refresh_vars_names_proc_rec_c NoRelevant new_vars_subst [] proc

let refresh_vars_names_proc side proc =
  let vars,names = get_vars_names_process proc.proc in
  let new_vars_subst = List.map (fun x -> x,fresh_var x.typ_v side, None) vars in
  let new_names_subst = List.map (fun n -> n,fresh_name n.typ_n) names in
  refresh_vars_names_proc_rec_c side new_vars_subst new_names_subst proc

(* Projection function: side = true <=> left, side = false <=> right *)
let rec proj_term side = function
  | Var v -> Var v
  | DumVar v -> DumVar v
  | Name n -> Name n
  | Sum(v,n) -> Sum(v,n)
  | Nat n -> Nat n
  | Fun(f,[t1;t2]) when List.exists (fun f' -> f == f') [choice_symb] ->
    if side then
      proj_term side t1
    else
      proj_term side t2
  | Fun(f,args) -> Fun(f,List.map (fun t -> proj_term side t) args)
  | Tuple(args) -> Tuple(List.map (fun t -> proj_term side t) args)
  | Restr(n,op,t) -> Restr(n,op,t)
  | Test(cond,t_then,t_else_op) ->
    begin
      match t_else_op with
      | None -> Test(proj_term side cond, proj_term side t_then, None)
      | Some t_else -> Test(proj_term side cond, proj_term side t_then, Some (proj_term side t_else))
    end
  | Let(pat,cond,t_then,t_else_op) ->
    begin
      match t_else_op with
      | None -> Let(proj_pattern side pat, proj_term side cond, proj_term side t_then, None)
      | Some t_else -> Let(proj_pattern side pat, proj_term side cond, proj_term side t_then, Some (proj_term side t_else))
    end
  | _ -> failwith "[transform.ml >> proj_term] Syntax not supported"

  and proj_pattern side = function
    | PatVar v -> PatVar v
    | PatEq t -> PatEq (proj_term side t)
    | PatTuple l -> PatTuple (List.map (fun p -> proj_pattern side p) l)
    | PatChoiceApp(p1,p2) ->
      if side then
        proj_pattern side p1
      else
        proj_pattern side p2
    | PatData(s,l) -> PatData(s,(List.map (fun p -> proj_pattern side p) l))


let rec contains typ = function
| Var _ -> false
| DumVar _ -> false
| Name _ -> false
| Sum(_,_) -> false
| Nat _ -> false
| Fun(f,args) ->
  begin
    match f.funcat with
    | x when List.mem x typ -> true
    | _ -> List.fold_left (fun b t -> b && (contains typ t)) true args
  end
| Tuple(args) -> List.fold_left (fun b t -> b && (contains typ t)) true args
| Restr(_,_,t) -> contains typ t
| Test(cond,t_then,t_else_op) -> (contains typ cond) && (contains typ t_then) && (match t_else_op with | None -> true | Some t_else -> contains typ t_else)
| _ -> failwith "[transform.ml >> contains_dest] Syntax not supported"

let is_well_formed = function
  | Fun(f,[t1;t2]) when f==eq_symb && not(contains [Logic;Dest] t1) && not(contains [Logic;Dest] t2) -> true
  | _ -> failwith "[transform.ml >> is_well_formed] Formulas can only be of the for 'u = v' with u,v two constructor terms"

let get_equal_terms t =
  assert(is_well_formed t);
  match t with
  | Fun(_,[t1;t2]) -> (t1,t2)
  | _ -> failwith "[transform.ml >> get_equal_terms] Assertion must have failed"

let rec create_new_events_inserted = function
  | [] -> []
  | DTable(tab)::q ->
    let new_event_inserted = {
    id_e =  { id = "Inserted_" ^ tab.id_t.id ; counter =0 };
    arg_ty_e = tab.arg_ty_t;
    side_e = NoRelevant;
    } in
    DTable(tab)::DEvent(new_event_inserted)
    ::(create_new_events_inserted q)
  | x::q -> x::(create_new_events_inserted q)

let create_badL_badR = function
  | l ->  DFree(badL) :: DFree(badR) :: DFree(c_error) :: l


let rec pattern_to_term = function
  | PatVar v -> DumVar v
  | PatEq _ -> failwith "[transform.ml >> pattern_to_term] Get command should not contain '=u' tests"
  | PatTuple l -> Tuple (List.map (fun p -> pattern_to_term p) l)
  | PatChoiceApp(p1,p2) -> Fun(choice_symb,[pattern_to_term p1; pattern_to_term p2])
  | PatData(s,l) -> Fun(s,(List.map (fun p -> pattern_to_term p) l))

let rec get_dum_var = function
  | PatVar v -> [v]
  | PatEq _ -> failwith "[transform.ml >> get_dum_var] Get command should not contain '=u' tests"
  | PatTuple l -> List.fold_left (fun acc pat -> List.append acc (get_dum_var pat)) [] l
  | PatChoiceApp(p1,p2) -> List.append (get_dum_var p1) (get_dum_var p2)
  | PatData(_,l) -> List.fold_left (fun acc pat -> List.append acc (get_dum_var pat)) [] l

let rec introduce_dum_var subst = function
  | Var x when List.mem x subst -> DumVar x
  | Var x -> Var x
  | DumVar x -> DumVar x
  | Name n -> Name n
  | Sum(v,n) -> Sum(v,n)
  | Nat n -> Nat n
  | Fun(f,args) ->
    Fun(f, List.map (fun t -> introduce_dum_var subst t) args)
  | Tuple(args) -> Tuple(List.map (fun t -> introduce_dum_var subst t) args)
  | Restr(n,op,t) -> Restr(n,op,t)
  | Test(cond,t_then,t_else_op) ->
      begin
        match t_else_op with
        | None -> Test(introduce_dum_var subst cond, introduce_dum_var subst t_then, None)
        | Some t_else -> Test(introduce_dum_var subst cond, introduce_dum_var subst t_then, Some (introduce_dum_var subst t_else))
      end
  | Let(pat,cond,t_then,t_else_op) ->
    begin
      match t_else_op with
      | None -> Let(pat, introduce_dum_var subst cond, introduce_dum_var subst t_then, None)
      | Some t_else -> Let(pat, introduce_dum_var subst cond, introduce_dum_var subst t_then, Some (introduce_dum_var subst t_else))
    end
  | _ -> failwith "[transform.ml >> introduce_dum_var] Syntax not supported"

let rec get_atoms = function
  | Var x -> [Var x]
  | DumVar x -> [DumVar x]
  | Name n -> [Name n]
  | Nat n -> [Nat n]
  | Fun(_,args) ->
    List.fold_left (fun acc t -> List.append acc (get_atoms t)) [] args
  | Tuple(args) -> List.fold_left (fun acc t -> List.append acc (get_atoms t)) [] args
  | Restr(n,op,t) -> [Restr(n,op,t)]
  | Test(cond,t_then,t_else_op) ->
      begin
        match t_else_op with
        | None -> List.append (get_atoms cond) (get_atoms t_then)
        | Some t_else -> List.append (get_atoms cond) (List.append (get_atoms t_then) (get_atoms t_else))
      end
  | Let(_,cond,t_then,t_else_op) ->
    begin
      match t_else_op with
      | None -> List.append (get_atoms cond) (get_atoms t_then)
      | Some t_else -> List.append (get_atoms cond) (List.append (get_atoms t_then) (get_atoms t_else))
    end
  | _ -> failwith "[transform.ml >> get_atoms] Syntax not supported"

let get_atoms_without_dum t =
  List.filter (fun x -> match x with | DumVar _ -> false | _ -> true) (get_atoms t)

(* Store the data needed to generate the axioms *)
let axiom_list = ref []

let rec remove_duplicates = function
  | [] -> []
  | h::t -> h::(remove_duplicates (List.filter (fun x -> x<>h) t))

let rec name_to_var = function
  | [] -> []
  | a::q -> match a with
    | Var x -> (Var x)::(name_to_var q)
    | DumVar x -> (DumVar x)::(name_to_var q)
    | Name n -> (Var { id_v = {id = "x" ^ n.id_n.id; counter = n.id_n.counter}; typ_v = n.typ_n; status_v = SCNone; options_v = []; side_v = NoRelevant })::(name_to_var q)
    | Restr(n,_,_) -> (Var { id_v = {id = "x" ^ n.id_n.id; counter = n.id_n.counter}; typ_v = n.typ_n; status_v = SCNone; options_v = []; side_v = NoRelevant })::(name_to_var q)
    | _ -> failwith "[transform.ml >> name_to_var] Call to name_to_var with a list of non-atomic terms"

let rec name_to_var_term = function
  | Var x -> List.hd (name_to_var [Var x])
  | DumVar x -> List.hd (name_to_var [DumVar x])
  | Name n -> List.hd (name_to_var [Name n])
  | Restr(n,op,t) -> List.hd (name_to_var [Restr(n,op,t)])
  | Fun(f,args) -> Fun(f,List.map (fun t -> name_to_var_term t) args)
  | Tuple(args) -> Tuple(List.map (fun t -> name_to_var_term t) args)
  | Test(cond,t_then,t_else_op) ->
      begin
        match t_else_op with
        | None -> Test(name_to_var_term cond, name_to_var_term t_then, None)
        | Some t_else -> Test(name_to_var_term cond, name_to_var_term t_then, Some (name_to_var_term t_else))
      end
  | Let(pat,cond,t_then,t_else_op) ->
    begin
      match t_else_op with
      | None -> Let(pat,name_to_var_term cond, name_to_var_term t_then, None)
      | Some t_else -> Let(pat,name_to_var_term cond, name_to_var_term t_then, Some (name_to_var_term t_else))
    end
  | _ -> failwith "[transform.ml >> name_to_var_term] Syntax not supported"

(* (Warning) - must be used carefully because it might generate bad-formed axioms... *)
let rec generate_axioms_aux = function
  | [] -> []
  | (evIns,evFail,pat_list,t1,t2)::q ->

    let atoms_t1_t2 = remove_duplicates (List.append (get_atoms_without_dum t1) (get_atoms_without_dum t2)) in
    let atoms_t1_t2' = name_to_var atoms_t1_t2 in
    let atoms_t1_t2'_var = List.map (fun t -> match t with | Var x -> x | DumVar x -> x | _ -> error_v) atoms_t1_t2' in

    let vars_pat_list = List.fold_left (fun acc pat -> List.append acc (get_vars_term pat)) [] pat_list in
    let vars_pat_list_subst = List.map (fun x -> (x,fresh_var x.typ_v Left, Some (fresh_var x.typ_v Right))) vars_pat_list in
    let new_pat_list = List.map (fun t -> refresh_vars_names_terms NoRelevant vars_pat_list_subst [] t) pat_list in
    let vars_new_pat_list = List.fold_left (fun acc pat -> List.append acc (get_vars_term pat)) [] new_pat_list in

    let list_of_var = List.append atoms_t1_t2'_var vars_new_pat_list in

    let subst_for_term =
      begin
        match evFail.side_e with
        | Left -> List.map (fun elt -> match elt with | (x,xL,Some _) -> (x,xL,None) | _ -> input_error (Printf.sprintf "[transform.ml >> generate_axioms_aux] vars_pat_list_subst has not the expected form") dummy_ext) vars_pat_list_subst
        | Right -> List.map (fun elt -> match elt with | (x,_,Some xR) -> (x,xR,None) | _ ->  input_error (Printf.sprintf "[transform.ml >> generate_axioms_aux] vars_pat_list_subst has not the expected form") dummy_ext) vars_pat_list_subst


        | NoRelevant -> []
      end
    in
    let right_hand_pred =
      if t1 = t2 then
        QFact(FPred(false_symb,[]))
      else
        QFact(FPred(neq_symb,[refresh_vars_names_terms NoRelevant subst_for_term [] (name_to_var_term t1);refresh_vars_names_terms NoRelevant subst_for_term [] (name_to_var_term t2) ]))
    in
    DEvent(evFail)::
    DAxiom(list_of_var, [FQuery(
    QQuery(
      QAnd(
        QFact(FEvent(evFail,atoms_t1_t2',false)),
        QFact(FEvent(evIns,new_pat_list,false))
      ),
      right_hand_pred
      )
    )])::(generate_axioms_aux q)

let rec generate_axioms = function
  | [] -> []
  | (DProcess proc_c) :: q -> List.append (generate_axioms_aux !axiom_list) ((DProcess proc_c) :: q)
  | x :: q -> x :: (generate_axioms q)

let count_get_duplicated = ref 0

let rec duplicate_get proc =
  match proc with
  | CNil -> CNil
  | CPar (p1,p2) -> CPar (duplicate_get_c p1, duplicate_get_c p2)
  | CRepl p -> CRepl (duplicate_get_c p)
  | CRestr (n, vars, p) -> CRestr (n, vars, duplicate_get_c p)
  | CTest (t, p_then, p_else) -> CTest (t, duplicate_get_c p_then, duplicate_get_c p_else)
  | CLet (pat,t,p_then,p_else) -> CLet (pat,t, duplicate_get_c p_then, duplicate_get_c p_else)
  | CInsert(tab,l,p) ->
    let event_inserted = {
      id_e =  { id = "Inserted_" ^ tab.id_t.id ; counter =0 };
      arg_ty_e = tab.arg_ty_t;
      side_e = NoRelevant;
    } in
    CEvent(
      event_inserted,
      l,
      None,
      { proc =  CInsert(tab,l,duplicate_get_c p); comment=None }
    )
  | CInput(c,pat,p) -> CInput(c,pat,duplicate_get_c p)
  | COutput(c,t,p) -> COutput(c,t, duplicate_get_c p)
  | CEvent(ev,args,vars,p) -> CEvent(ev,args,vars,duplicate_get_c p)

  | CGet(tab, pat_list, phi, t_then, t_else) ->
    begin
      match phi with
      | None ->
        let event_inserted = {
          id_e =  { id = "Inserted_" ^ tab.id_t.id ; counter =0 };
          arg_ty_e = tab.arg_ty_t;
          side_e = NoRelevant;
        } in
        count_get_duplicated := !count_get_duplicated + 1;
        let event_failLeft = {
        id_e =  { id = "Fail_" ^ (string_of_int !count_get_duplicated) ; counter =0 };
        arg_ty_e = [];
        side_e = Left;
        } in
        count_get_duplicated := !count_get_duplicated + 1;
        let event_failRight_1 = {
        id_e =  { id = "Fail_" ^ (string_of_int !count_get_duplicated) ; counter =0 };
        arg_ty_e = [];
        side_e = Right;
        } in

        count_get_duplicated := !count_get_duplicated + 1;
        let event_failRight_2 = {
        id_e =  { id = "Fail_" ^ (string_of_int !count_get_duplicated) ; counter =0 };
        arg_ty_e = [];
        side_e = Right;
        } in

        let c_out_bad = { proc = COutput(Name c_error, Fun(choice_symb,[Name badL; Name badR]), { proc = CNil; comment=None}) ; comment=None } in

        let t_thenthen = duplicate_get_c t_then in
        let t_thenelse = duplicate_get_c { proc = CEvent(event_failRight_1, [], None, c_out_bad) ; comment=None } in

        let t_elseboth =
        {
        proc = CEvent(event_failLeft, [], None,

        refresh_vars_names_proc Right {
        proc =
          CGet(
            tab,
            pat_list,
            None,
            c_out_bad,
            { proc = CEvent(event_failRight_2, List.fold_left (fun acc pat -> List.append acc (get_atoms_without_dum (pattern_to_term (proj_pattern false pat)))) [] pat_list, None, duplicate_get_c t_else); comment=None }
            );
        comment=None }

        ) ; comment = None
        } in

        axiom_list := (event_inserted, event_failRight_2, List.map (fun pat -> proj_term false (pattern_to_term pat)) pat_list, Fun(false_symb,[]) , Fun(false_symb,[]))::!axiom_list;
        axiom_list := (event_inserted, event_failRight_1, List.map (fun pat -> proj_term false (pattern_to_term pat)) pat_list, Fun(false_symb,[]) , Fun(false_symb,[]))::!axiom_list;
        axiom_list := (event_inserted, event_failLeft, List.map (fun pat -> proj_term true (pattern_to_term pat)) pat_list, Fun(false_symb,[]), Fun(false_symb,[]))::!axiom_list;

        CGet(
          tab,
          pat_list,
          None,
          refresh_vars_names_proc Right { proc = CGet(tab, pat_list, None, t_thenthen, t_thenelse); comment=None },
          { proc = t_elseboth.proc; comment=None }
          )
      | Some formula when is_well_formed formula ->
        let dum_var_in_pat = List.fold_left (fun acc t -> List.append acc (get_dum_var t)) [] pat_list in

        let u_then,u_else = get_equal_terms formula in
        let atoms_t1_t2_left = remove_duplicates (List.append (get_atoms_without_dum (introduce_dum_var dum_var_in_pat (proj_term true u_then))) (get_atoms_without_dum (introduce_dum_var dum_var_in_pat (proj_term true u_else)))) in
        let atoms_t1_t2_right = remove_duplicates (List.append (get_atoms_without_dum (introduce_dum_var dum_var_in_pat (proj_term false u_then))) (get_atoms_without_dum (introduce_dum_var dum_var_in_pat (proj_term false u_else)))) in
        let event_inserted = {
          id_e =  { id = "Inserted_" ^ tab.id_t.id ; counter =0 };
          arg_ty_e = tab.arg_ty_t;
          side_e = NoRelevant;
        } in
        count_get_duplicated := !count_get_duplicated + 1;
        let event_failLeft = {
        id_e =  { id = "Fail_" ^ (string_of_int !count_get_duplicated) ; counter =0 };
        arg_ty_e = List.map (fun t -> match t with Var x -> x.typ_v | Name n -> n.typ_n | Restr(n,_,_) -> n.typ_n | _ -> bitstring_id) atoms_t1_t2_left; (* left and right have the same types; bitstring_id is by default *)
        side_e = Left;
        } in
        count_get_duplicated := !count_get_duplicated + 1;
        let event_failRight_1 = {
        id_e =  { id = "Fail_" ^ (string_of_int !count_get_duplicated) ; counter =0 };
        arg_ty_e = List.map (fun t -> match t with Var x -> x.typ_v | Name n -> n.typ_n | Restr(n,_,_) -> n.typ_n | _ -> bitstring_id) atoms_t1_t2_left; (* left and right have the same types; bitstring_id is by default *)
        side_e = Right;
        } in

        count_get_duplicated := !count_get_duplicated + 1;
        let event_failRight_2 = {
        id_e =  { id = "Fail_" ^ (string_of_int !count_get_duplicated) ; counter =0 };
        arg_ty_e = List.map (fun t -> match t with Var x -> x.typ_v | Name n -> n.typ_n | Restr(n,_,_) -> n.typ_n | _ -> bitstring_id) atoms_t1_t2_left; (* left and right have the same types; bitstring_id is by default *)
        side_e = Right;
        } in

        let c_out_bad = { proc = COutput(Name c_error, Fun(choice_symb,[Name badL; Name badR]), { proc = CNil; comment=None}) ; comment=None } in

        let t_thenthen = duplicate_get_c t_then in
        let t_thenelse = duplicate_get_c { proc = CEvent(event_failRight_1, atoms_t1_t2_right, None, c_out_bad) ; comment=None } in

        let t_elseboth =
        {
        proc = CEvent(event_failLeft, atoms_t1_t2_left, None,

        refresh_vars_names_proc Right {
        proc =
          CGet(
            tab,
            pat_list,
            Some (proj_term false formula),
            c_out_bad,
            { proc = CEvent(event_failRight_2, remove_duplicates (List.append (get_atoms_without_dum (introduce_dum_var dum_var_in_pat (proj_term false u_then))) (get_atoms_without_dum (introduce_dum_var dum_var_in_pat (proj_term false u_else)))), None, duplicate_get_c t_else); comment=None }
            );
        comment=None }

        ) ; comment = None
        } in

        axiom_list := (event_inserted, event_failRight_2, List.map (fun pat -> proj_term false (pattern_to_term pat)) pat_list, introduce_dum_var dum_var_in_pat (proj_term false u_then), introduce_dum_var dum_var_in_pat (proj_term false u_else))::!axiom_list;
        axiom_list := (event_inserted, event_failRight_1, List.map (fun pat -> proj_term false (pattern_to_term pat)) pat_list, introduce_dum_var dum_var_in_pat (proj_term false u_then), introduce_dum_var dum_var_in_pat (proj_term false u_else))::!axiom_list;
        axiom_list := (event_inserted, event_failLeft, List.map (fun pat -> proj_term true (pattern_to_term pat)) pat_list, introduce_dum_var dum_var_in_pat (proj_term true u_then), introduce_dum_var dum_var_in_pat (proj_term true u_else))::!axiom_list;

        CGet(
          tab,
          pat_list,
          Some (proj_term true formula),
          refresh_vars_names_proc Right { proc = CGet(tab, pat_list, Some (proj_term false formula), t_thenthen, t_thenelse); comment=None },
          { proc = t_elseboth.proc; comment=None }
          )
      | _ -> input_error (Printf.sprintf "[transform.ml >> duplicate_get] Formulas can only be of the for 'u = v' with u,v two constructor terms") dummy_ext
    end
  | CPhase(i,p) -> CPhase(i, duplicate_get_c p)
  | CLetFilter _ -> failwith "[transform.ml >> duplicate_get] Here"


and duplicate_get_c proc = { proc = duplicate_get proc.proc; comment=None }

let apply_transform_duplicate_get decl =
  match decl with
  | DProcess proc_c -> DProcess (duplicate_get_c (refresh_vars_proc_choice proc_c))
  | _ -> decl


(*******************************
***         Parsing          ***
********************************)

let parse filename =
  try
    let ic = open_in filename in
    let lexbuf = Lexing.from_channel ic in
    lexbuf.Lexing.lex_curr_p <- { lexbuf.Lexing.lex_curr_p with
                                  Lexing.pos_fname = filename };
    let ptree =
      try
        Pitparser.all Pitlexer.token lexbuf
      with Parsing.Parse_error ->
        input_error "Syntax error" (extent lexbuf)
    in
    close_in ic;
    ptree
  with Sys_error s ->
    user_error ("File error: " ^ s ^ "\n")

let parse_file s =
  Printf.printf "Parsing the file...\n";
  let (decl_l, proc, _) = parse s in

  (*if second_proc <> None
  then user_error "Equivalence properties are not handled yet by the front-end.";*)

  let decl1 = expand_tdecl initial_environment proc decl_l in

  ignore_type := true;

  set_ignore_type decl1;

  let decl2 = create_badL_badR decl1 in

  allowDiffPatterns := true;
  let decl3 = set_allowDiffPatterns decl2 in

  reconstructTrace := false;
  let decl4 = set_reconstructTrace decl3 in

  let decl5 = create_new_events_inserted decl4 in

  let decl6 = List.map (fun x -> apply_transform_duplicate_get x) decl5 in

  let decl7 = generate_axioms decl6 in

  decl7
